import calendar
from datetime import datetime, timedelta
from typing import Optional
from zoneinfo import ZoneInfo

import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_discord.container import Guild
from cpl_query.extension import List
from discord import PrivacyLevel
from discord.scheduled_event import ScheduledEvent as DiscordEvent

from bot_data.abc.scheduled_event_repository_abc import ScheduledEventRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.scheduled_event import ScheduledEvent
from bot_data.model.scheduled_event_interval_enum import ScheduledEventIntervalEnum
from bot_data.model.server_config import ServerConfig
from modules.base.model.active_event import ActiveEvent


class EventService:
    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        servers: ServerRepositoryABC,
        users: UserRepositoryABC,
        db: DatabaseContextABC,
        events: ScheduledEventRepositoryABC,
    ):
        self._config = config
        self._logger = logger
        self._servers = servers
        self._users = users
        self._db = db
        self._events = events

        self._active_events = List(ActiveEvent)

    def add_event(self, event: ActiveEvent):
        if self._active_events.contains(event):
            return

        self._active_events.add(event)

    def get_active_event(self, event: discord.ScheduledEvent) -> Optional[ActiveEvent]:
        return self._active_events.where(lambda x: x.event.id == event.id).single_or_default()

    def get_active_event_by_channel_id(self, channel_id: int) -> Optional[ActiveEvent]:
        return self._active_events.where(
            lambda x: x.event.channel is not None and x.event.channel.id == channel_id
        ).single_or_default()

    def remove_event(self, event: ActiveEvent):
        if not self._active_events.contains(event):
            return

        self._active_events.remove(event)

    def give_xp_for_event_participation(self, member: discord.Member, active_event: ActiveEvent):
        server = self._servers.get_server_by_discord_id(member.guild.id)
        user = self._users.get_user_by_discord_id_and_server_id(member.id, server.id)
        if active_event.participants.any(lambda x: x.id == user.id):
            self._logger.debug(__name__, f"Module {type(self)} stopped")
            return

        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{server.discord_id}")
        user.xp += settings.xp_per_event_participation
        self._users.update_user(user)
        self._db.save_changes()
        active_event.participants.append(user)

    def _append_interval(self, interval: ScheduledEventIntervalEnum, ts: datetime) -> datetime:
        if ts >= datetime.now():
            return ts

        if interval == ScheduledEventIntervalEnum.daily:
            ts = ts + timedelta(days=1)

        elif interval == ScheduledEventIntervalEnum.weekly:
            ts = ts + timedelta(weeks=1)

        elif interval == ScheduledEventIntervalEnum.monthly:
            days_in_month = calendar.monthrange(ts.year, ts.month + 1)[1]
            ts = ts + timedelta(days=days_in_month)

        elif interval == ScheduledEventIntervalEnum.yearly:
            ts = ts + timedelta(days=365)

        while ts < datetime.now():
            ts = self._append_interval(interval, ts)

        return ts

    async def check_and_create_scheduled_events(self, guild: Guild):
        server = self._servers.get_server_by_discord_id(guild.id)
        scheduled_events_from_db = self._events.get_scheduled_events_by_server_id(server.id)
        for scheduled_event in scheduled_events_from_db:
            scheduled_event: ScheduledEvent = scheduled_event
            from_guild = List(DiscordEvent, guild.scheduled_events).where(
                lambda x: x.name == scheduled_event.name
                and x.description == scheduled_event.description
                and x.entity_type == scheduled_event.entity_type
            )
            if from_guild.count() != 0:
                continue

            kwargs = {"name": scheduled_event.name, "description": scheduled_event.description}

            if scheduled_event.channel_id is not None:
                kwargs["channel"] = guild.get_channel(scheduled_event.channel_id)

            if scheduled_event.start_time is not None:
                scheduled_event.start_time = self._append_interval(scheduled_event.interval, scheduled_event.start_time)

                start_time = scheduled_event.start_time.replace(tzinfo=ZoneInfo("Europe/Berlin"))

                kwargs["start_time"] = start_time

            if scheduled_event.end_time is not None:
                scheduled_event.end_time = self._append_interval(scheduled_event.interval, scheduled_event.end_time)
                end_time = scheduled_event.end_time.replace(tzinfo=ZoneInfo("Europe/Berlin"))
                kwargs["end_time"] = end_time

            kwargs["entity_type"] = scheduled_event.entity_type
            if scheduled_event.location is not None:
                kwargs["location"] = scheduled_event.location

            kwargs["privacy_level"] = PrivacyLevel.guild_only

            try:
                self._logger.debug(__name__, f"Try to create scheduled event for guild {guild.name}")
                await guild.create_scheduled_event(**kwargs)
                self._events.update_scheduled_event(scheduled_event)
                self._db.save_changes()
            except Exception as e:
                self._logger.error(__name__, f"Watching scheduled events failed", e)
